<?php

namespace Drp\LaravelSetup\Database;

use Drp\LaravelSetup\Database\Env\Writer;
use Illuminate\Console\Command as LaravelCommand;
use Illuminate\Database\Connection;
use Illuminate\Database\Connectors\ConnectionFactory;
use Illuminate\Database\QueryException;

class CreateDatabase
{
    /**
     * @var LaravelCommand
     */
    private $output;

    /**
     * @var Writer
     */
    private $envWriter;

    public function __construct(Writer $envWriter)
    {
        $this->envWriter = $envWriter;
    }

    public function execute(string $defaultDbName): bool
    {
        $config = $this->requestConnectionSettings($defaultDbName);

        if ($config === false) {
            $this->output->error('Failed to get valid connection settings, skipping...');

            return false;
        }

        $this->createDatabase($config);

        $this->output->getOutput()->success('Writing credentials to .env file');
        $this->envWriter->put($config);

        return true;
    }

    public function getCommand(): string
    {
        return 'Database setup';
    }

    public function setOutput(LaravelCommand $command): self
    {
        $this->output = $command;

        return $this;
    }

    private function wantsDatabaseCreating(): bool
    {
        return $this->output->confirm(
            'Would you like to create the database now?',
            true
        );
    }

    private function failedToConnect(array $config): bool
    {
        $verifier = app(ConnectionVerifier::class);
        $connectionConfig = array_merge($config, ['database' => '']);

        if ($verifier($connectionConfig) === false) {
            $this->output->getOutput()->block(
                [
                    'Failed to make a successful connection.' . PHP_EOL . PHP_EOL .
                    $verifier->getLastError() . PHP_EOL,
                ],
                'ERR',
                'error',
                '  ',
                true
            );

            return true;
        }

        return false;
    }

    /**
     * @return bool|array
     */
    private function requestConnectionSettings(string $defaultDbName)
    {
        $config = ConnectionSettings::request($this->output, $defaultDbName);

        if ($this->failedToConnect($config)) {
            if ($this->output->confirm('Would you like trying to try again?', true)) {
                return $this->requestConnectionSettings($defaultDbName);
            }

            return false;
        }

        return $config;
    }

    private function createDatabase(array $config): bool
    {
        $connection = $this->makeConnection($config);
        $databaseName = $config['database'];

        if ($this->databaseDoesNotExist($connection, $databaseName) === false) {
            $this->output->alert('Detected database already exists, skipping creating the database.');

            return true;
        }

        if ($this->wantsDatabaseCreating()) {
            $this->makeDatabase($connection, $databaseName);
        }

        if ($this->databaseDoesNotExist($connection, $databaseName)) {
            if ($this->output->confirm('Could not find database, would you like to skip creating a database?')) {
                return false;
            }

            return $this->createDatabase($config);
        }

        return true;
    }

    private function makeConnection(array $config): Connection
    {
        /** @var ConnectionFactory $connectionFactory */
        $connectionFactory = app(ConnectionFactory::class);
        $connectionConfig = array_merge($config, ['database' => '']);

        return $connectionFactory->make($connectionConfig);
    }

    private function databaseDoesNotExist(Connection $connection, string $database): bool
    {
        $dbResult = $connection->select('SHOW DATABASES');
        // Cannot use parameter binding for db, table or column names
        $result = collect($dbResult)->pluck('Database');

        return in_array($database, $result->toArray(), false) === false;
    }

    private function makeDatabase(Connection $connection, string $database): bool
    {
        try {
            // Cannot use parameter binding for db, table or column names
            $connection->statement("CREATE DATABASE `{$database}`");
        } catch (QueryException $exception) {
            return false;
        }

        return true;
    }
}
